// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include  "MZO.h"

#include <math.h>

const string MZO::_esa("esa");
const string MZO::_pdist("pdist");
const string MZO::_p0("p0");
const string MZO::_ptop("ptop");
const string MZO::_pdelta("pdelta");
const string MZO::_powtop("powtop");
const string MZO::_normtop("normtop");

MZO::MZO( const string& id, const string& hduname, RDB& rdb ) throw ( Exception ) : MZ(id) { 

  TraceFct tf("MZO::MZO"); 

  _zmin = rdb.getDataDouble( "zmin" );
  _zmax = rdb.getDataDouble( "zmax" );

  init( hduname ); 

}

bool MZO::lookUpScatterAngle( double energy, double sin_alpha, double rannum, bool inplane, double& theta ) {

#ifdef DEBUG  
  TraceFct tf( "MZO::lookUpScatterAngle" );
#endif

  double esa = energy * sin_alpha;

  double *esa_lo, *esa_hi;

  bool is_extrapolated;

  if ( NULL != bnd_bsearch( &esa, _data+_offsets[_esa], _nrows, _rowsize, &esa_lo, &esa_hi, esacmp ) ) {            // Exact match!!!

    is_extrapolated = false;

  } else if ( NULL == esa_lo ) { // Less than all _esa

    esa_lo  = esa_hi;
    is_extrapolated = true;

  } else if ( NULL == esa_hi ) { // Greater than all _esa

    esa_hi  = esa_lo;
    is_extrapolated = true;

  } else {                       // Within _esa

    is_extrapolated = false;
    
  }

  long idx_lo = ( (char*)esa_lo - _data ) / _rowsize;
  long idx_hi = ( (char*)esa_hi - _data ) / _rowsize;

  rannum = 2.0 * ( rannum - 0.5 );
  bool is_neg = ( rannum < 0.0 );

  if ( is_neg ) 
    rannum = -1.0 * rannum;

  double theta_esa_lo = lookUpTheta( idx_lo, rannum );
  double theta_esa_hi = lookUpTheta( idx_hi, rannum );


  if ( theta_esa_lo < 0.0 ) theta_esa_lo = 0.0;
  if ( theta_esa_hi < 0.0 ) theta_esa_hi = 0.0;

  theta = interpolate( esa, *esa_lo, theta_esa_lo, *esa_hi, theta_esa_hi );
  
  if ( is_neg )
    theta *= -1;

  if ( ! inplane )
    theta *= sin_alpha;

#ifdef DEBUG  
  if ( dbflag("theta" )) 
    tf.message( "p = %.14g theta = %.5g\n", rannum, theta );
#endif

  return is_extrapolated;
}

double MZO::lookUpTheta( const int idx, double rannum ) {

  if ( rannum <= p0( idx ) ) {

    return 0.0;

  } else if ( rannum <= ptop( idx ) ) {

    rannum = ( rannum - p0( idx ) ) / pdelta( idx );
    int jdx = (int) rannum;
    int ndist = _ns[_pdist];
    if ( jdx >= ndist )
      jdx = ndist - 1;
    rannum = rannum - (double) jdx;
    double t = (pdist( idx ))[jdx];

    return ( rannum * (pdist( idx ))[jdx+1] + ( 1.0 - rannum ) * t );

  } else {

    double rtop = 0.99999; // Ouch!  this was in the old code...

    if ( rannum < rtop ) {

      return pow( normtop( idx ) * ( 1.0 - rannum ), powtop( idx ) );

    } else {

      return pow( normtop( idx ) * ( 1.0 - rtop ), powtop( idx ) );

    }

  }

}

void MZO::init( const string& hduname ) throw ( Exception ) {
  TraceFct tf( "MZO::init" );

  Fun fun = FunOpen( const_cast<char*>(hduname.c_str()),
		     const_cast<char*>("r"), NULL );
  if ( ! fun )
    throw( Exception( string( "error opening scattering distribution table '")
		      + hduname + "'" ) );

  try { 

    verifyCards(   fun );

    vector<string> names;
    names.push_back( _esa      );
    names.push_back( _pdist   ); 
    names.push_back( _p0      ); 
    names.push_back( _ptop    ); 
    names.push_back( _pdelta  ); 
    names.push_back( _powtop  ); 
    names.push_back( _normtop );
    verifyColumns( fun, names );

    verifyData(    fun );

  } catch ( Exception& E ) {

    throw( E );

  }

  FunClose( fun );
}

